####################################################################################################################
## Calculate air temperature and pressure daily means using ERA5-Land (or, in alternative, a MeteoCH station - for temperature only)
####################################################################################################################
## Input and output are in Station Exchange Format (SEF), except MeteoCH data which are in their original format.

rm(list=ls())

require(dataresqc)
require(geosphere)
require(ncdf4)
require(XLConnect)

main <- "/anydir/"
setwd(main)

# read inventory
inventory <- readWorksheetFromFile(paste0(main,"code/data_input/arm_stations_inventory_withnbcn.xlsx"),sheet=1, header=T,colTypes="character",startRow=2)

sef <- "code/data_prep/qced_subdaily/"
cor <- "code/data_prep/daymeans/meteoCH/"
era5_t <- "code/data_prep/daymeans/ERA5-Land/reanalysis-era5-land_2m_temperature_daily_cycle_1981-2010.nc"
era5_p <- "code/data_prep/daymeans/ERA5-Land/reanalysis-era5-land_surface_pressure_daily_cycle_1981-2010.nc"
outpath <- "code/data_prep/daymeans/sef/"

########################
## Temperature
########################
subdaily_ta <- inventory[which(inventory[,"var"]=="ta" & inventory[,"resolution"]=="subdaily"),"subdaily_file"]
fnames <- lapply(inventory[inventory[,"var"]=="ta" & inventory[,"resolution"]=="subdaily","orig_filename"],
                 function(x){unlist(strsplit(x,","))})

seffiles <- list.files(sef)
seffiles_ta <- seffiles[grepl("_ta.tsv|_ta_",seffiles)]

stn <- seffiles_ta[4]

for(stn in seffiles_ta){
  print(stn)
  addvar <- gsub(".tsv","",gsub(".*_ta_|.*_ta.tsv","", stn))
  new_note <- paste0(addvar,ifelse(nchar(addvar)>0,"_",""),"daily")
  ## select the correct mch stations
  meteoCH <- subdaily_ta[which(sapply(fnames, function(x){any(x%in%stn)}))]
  if(length(meteoCH)>1){meteoCH <- meteoCH[1]}
  ## Temperature
  if(length(meteoCH)>0){
    if (meteoCH == "ERA-5") {
      nct <- nc_open(era5_t)
      lons <- ncvar_get(nct, "longitude")
      lons[lons>180] <- lons[lons>180] - 360
      lats <- ncvar_get(nct, "latitude")
      coords <- data.frame(lon=rep(lons,each=length(lats)), lat=rep(lats,times=length(lons)))
    } 
    
    x <- read_sef(paste0(sef,stn), all=TRUE)
    x <- x[which(!is.na(x$Value)), ]
    ko <- grep("qc", x$Meta)
    if (length(ko) > 0) x <- x[-ko, ] # Remove flagged values
    meta <- read_meta(paste0(sef,stn))
    x$Time <- x$Hour + x$Minute/60
    ntimes <- aggregate(x$Time, list(x$Year,x$Month,x$Day), function(x) sum(!is.na(x))) # how many time steps per day
    ntimes <- ntimes[order(ntimes[,1],ntimes[,2],ntimes[,3]), ]
    out <- data.frame(Year=ntimes$Group.1,
                      Month=ntimes$Group.2,
                      Day=ntimes$Group.3,
                      Hour=rep(NA, nrow(ntimes)),
                      Minute=rep(NA, nrow(ntimes)),
                      Value=rep(NA, nrow(ntimes)))
    
    xlon <- as.numeric(meta["lon"])
    xlat <- as.numeric(meta["lat"])
    
    if (max(ntimes$x) > 0) {    
      x$Value[which(is.na(x$Time))] <- NA
      if (meteoCH == "ERA-5") {
        ## Select closest grid point on land (limit the search to a square of 25 grid points)       
        dc <- NA
        k <- 0
        coords_st <- coords
        while(is.na(dc[1]) & k<25) {
          i_grid <- which.min(distGeo(c(xlon,xlat), coords_st))
          i_lon <- which(lons == coords_st$lon[i_grid])
          i_lat <- which(lats == coords_st$lat[i_grid])
          ## Get daily cycle
          dc <- ncvar_get(nct, "t2m", c(i_lon,i_lat,1), c(1,1,-1))
          coords_st[i_grid,] <- NA
          k <- k + 1
        }
        if (k == 25) { ## if no land is found nearby the station, no correction is applied
          dc <- rep(0, 12*24)
          warning(paste("No land grid points close to", meta["id"]))
        }
        ## Convert to anomalies from the daily mean
        for (m in 1:12) dc[((m-1)*24+1):(m*24)] <- 
          dc[((m-1)*24+1):(m*24)] - mean(dc[((m-1)*24+1):(m*24)]) 
        dc <- append(dc, dc[1])
      } else {
        dc <- read.csv(paste0(cor,meteoCH), sep=";", na.strings="-")
        dc$m <- as.integer(substr(dc[,2],5,6)) # months
        dc$h <- as.integer(substr(dc[,2],9,10)) + as.integer(substr(dc[,2],11,12)) / 60
        dc <- aggregate(dc[,3], list(dc$h,dc$m), mean, na.rm=TRUE)$x ## ten minutes monthly mean
        # 144 is the number of measurements if every ten minutes is measured
        # the following generates the anomalie for the 144 time steps with respect to the monthly daily mean
        for (m in 1:12) dc[((m-1)*144+1):(m*144)] <- 
          dc[((m-1)*144+1):(m*144)] - mean(dc[((m-1)*144+1):(m*144)]) 
        dc <- append(dc, dc[1])
      }
      ## Calculate daily means
      xmean <- aggregate(x$Value, list(x$Year,x$Month,x$Day), mean, na.rm=TRUE)
      timeagg <- aggregate(x$Meta, list(x$Year,x$Month,x$Day), function(y){tim <- paste0(gsub(".*orig.time=","",y),collapse = "-")})
      xmean <- xmean[order(xmean$Group.1, xmean$Group.2, xmean$Group.3), ]
      timeagg <- timeagg[order(timeagg$Group.1, timeagg$Group.2, timeagg$Group.3), ]
      
      ## Apply correction for daily cycle
      for (i_day in 1:nrow(xmean)) {
        xtimes <- x$Time[which(x$Year==xmean$Group.1[i_day] & 
                                 x$Month==xmean$Group.2[i_day] & 
                                 x$Day==xmean$Group.3[i_day])]
        xtimes <- xtimes[!is.na(xtimes)]
        if (length(xtimes) > 0) {
          m <- as.integer(xmean$Group.2[i_day])
          if (meteoCH=="ERA-5") {
            dc_day <- dc[((m-1)*24+1):(m*24)]
            dc_day <- append(dc_day, dc_day[1])
            corrections <- dc_day[as.integer(xtimes)+1] * (1-xtimes+as.integer(xtimes)) + dc_day[as.integer(xtimes)+2] * (xtimes-as.integer(xtimes))
          } else {
            dc_day <- dc[((m-1)*144+1):(m*144)]
            dc_day <- append(dc_day, dc_day[1])
            corrections <- array(dim=length(xtimes))
            # get values from dc that correspond to time when is measured
            for (i_times in 1:length(xtimes)) {
              j <- which.min(abs(xtimes[i_times] - seq(0,24,length.out=145)))
              corrections[i_times] <- dc_day[j]
            }
          }
          out$Value[i_day] <- round(xmean$x[i_day] - mean(corrections), 1)
        }
      }
      write_sef(out, outpath, meta["var"], meta["id"], meta["name"], 
                meta["lat"], meta["lon"], meta["alt"], meta["source"], meta["link"], 
                meta["units"], "mean", period="day", metaHead= paste0(meta["meta"],"|Diurnal Cycle=Y"), note=new_note, keep_na=TRUE)
      
    } else if (length(ntimes$x)!=length(x[,"Value"])){
      print("no correction of daily cycle due to missing time, but still calculation of mean")
      outhelp <- aggregate(x$Value, list(x$Year,x$Month,x$Day), mean)
      outhelp <- outhelp[order(outhelp$Group.1,outhelp$Group.2,outhelp$Group.3),]
      out$Value <- outhelp$x
      write_sef(out, outpath, meta["var"], meta["id"], meta["name"], 
                meta["lat"], meta["lon"], meta["alt"], meta["source"], meta["link"], 
                meta["units"], "mean", period="day", metaHead= paste0(meta["meta"],"|Diurnal Cycle=N"), note=new_note, keep_na=TRUE)
    } else {
      out$Value <- x[,"Value"]
      write_sef(out, outpath, meta["var"], meta["id"], meta["name"], 
                meta["lat"], meta["lon"], meta["alt"], meta["source"], meta["link"], 
                meta["units"], "mean", period="day", metaHead= meta["meta"], note=new_note, keep_na=TRUE)
    }
    
    
    if (meteoCH == "ERA-5") nc_close(nct)
    
  } else {
    ## for daily files set to zero, when qc not passed
    x <- read_sef(paste0(sef,stn), all=TRUE)
    x <- x[which(!is.na(x$Value)), ]
    ko <- grep("qc", x$Meta)
    if (length(ko) > 0) x <- x[-ko, ] # Remove flagged values
    meta <- read_meta(paste0(sef,stn))
    x$Time <- x$Hour + x$Minute/60
    ntimes <- aggregate(x$Time, list(x$Year,x$Month,x$Day), function(x) sum(!is.na(x))) # how many time steps per day
    ntimes <- ntimes[order(ntimes[,1],ntimes[,2],ntimes[,3]), ]
    out <- data.frame(Year=ntimes$Group.1,
                      Month=ntimes$Group.2,
                      Day=ntimes$Group.3,
                      Hour=rep(24, nrow(ntimes)),
                      Minute=rep(0, nrow(ntimes)),
                      Value=x["Value"])
    
    write_sef(out, outpath, meta["var"], meta["id"], meta["name"], 
              meta["lat"], meta["lon"], meta["alt"], meta["source"], meta["link"], 
              meta["units"], "mean", period="day", meta["meta"], note=new_note, keep_na=TRUE)
  
  }
 }

########################
## Pressure
########################

seffiles <- list.files(sef)
seffiles_p <- seffiles[grepl("_p.tsv|_p_",seffiles)]

## all based on era-5
## for Tavel, no correction of the daily cycle is applied

ncp <- nc_open(era5_p)
lons <- ncvar_get(ncp, "longitude")
lons[lons>180] <- lons[lons>180] - 360
lats <- ncvar_get(ncp, "latitude")
coords <- data.frame(lon=rep(lons,each=length(lats)), lat=rep(lats,times=length(lons)))

stn <- seffiles_p[grepl("Tavel",seffiles_p)]
stn <- seffiles_p[1]

for (stn in seffiles_p) {
  print(stn)
  addvar <- gsub(".tsv","",gsub(".*_p_|.*_p.tsv","", stn))
  new_note <- paste0(addvar,ifelse(nchar(addvar)>0,"_",""),"daily")
  
  x <- read_sef(paste0(sef,stn), all=TRUE)
  x <- x[which(!is.na(x$Value)), ]
  ko <- grep("qc", x$Meta)
  if (length(ko) > 0) x <- x[-ko, ] # Remove flagged values
  if(grepl("Herisau_Nef",stn)){
    ko <- grep("PTC=N", x$Meta)
    if (length(ko) > 0) x <- x[-ko, ] # Remove if partly PTC=N
  }
  
  meta <- read_meta(paste0(sef,stn))
  x$Time <- x$Hour + x$Minute/60
  ntimes <- aggregate(x$Time, list(x$Year,x$Month,x$Day), function(x) sum(!is.na(x)))
  ntimes <- ntimes[order(ntimes[,1],ntimes[,2],ntimes[,3]), ]
  out <- data.frame(Year=ntimes$Group.1,
                    Month=ntimes$Group.2,
                    Day=ntimes$Group.3,
                    Hour=rep(NA, nrow(ntimes)),
                    Minute=rep(NA, nrow(ntimes)),
                    Value=rep(NA, nrow(ntimes)))
  xlon <- as.numeric(meta["lon"])
  xlat <- as.numeric(meta["lat"])
  
  if (max(ntimes$x) > 0) {
    ## Select closest grid point on land (limit the search to a square of 25 grid points)       
    dc <- NA
    k <- 0
    coords_st <- coords
    while(is.na(dc[1]) & k<25) {
      i_grid <- which.min(distGeo(c(xlon,xlat), coords_st))
      i_lon <- which(lons == coords_st$lon[i_grid])
      i_lat <- which(lats == coords_st$lat[i_grid])
      ## Get daily cycle
      dc <- ncvar_get(ncp, "sp", c(i_lon,i_lat,1), c(1,1,-1)) / 100
      coords_st[i_grid,] <- NA
      k <- k + 1
    }
    if (k == 25) { ## if no land is found nearby the station, no correction is applied
      dc <- rep(0, 12*24)
      warning(paste("No land grid points close to", meta["id"]))
    }
    ## Convert to anomalies from the daily mean
    for (m in 1:12) dc[((m-1)*24+1):(m*24)] <- 
      dc[((m-1)*24+1):(m*24)] - mean(dc[((m-1)*24+1):(m*24)]) 
    dc <- append(dc, dc[1])
    ## Calculate daily means
    xmean <- aggregate(x$Value, list(x$Year,x$Month,x$Day), mean, na.rm=TRUE)
    xmean <- xmean[order(xmean$Group.1, xmean$Group.2, xmean$Group.3), ]
    ## Apply correction for daily cycle (if all times are known)
    for (i_day in 1:nrow(xmean)) {
      xtimes <- x$Time[which(x$Year==xmean$Group.1[i_day] & 
                               x$Month==xmean$Group.2[i_day] & 
                               x$Day==xmean$Group.3[i_day])]
      if (sum(is.na(xtimes)) == 0) {
        m <- as.integer(xmean$Group.2[i_day])
        dc_day <- dc[((m-1)*24+1):(m*24)]
        dc_day <- append(dc_day, dc_day[1])
        corrections <- dc_day[as.integer(xtimes)+1] * (1-xtimes+as.integer(xtimes)) + 
          dc_day[as.integer(xtimes)+2] * (xtimes-as.integer(xtimes))
        out$Value[i_day] <- round(xmean$x[i_day] - mean(corrections), 1)
      } else {
        out$Value[i_day] <- round(xmean$x[i_day], 1)
      }
    }
    metahelp <- "|Diurnal Cycle=Y"
    
  } else if (length(ntimes$x)!=length(x[,"Value"])){
    print("no correction of daily cycle due to missing time, but still calculation of mean")
    outhelp <- aggregate(x$Value, list(x$Year,x$Month,x$Day), mean)
    outhelp <- outhelp[order(outhelp$Group.1,outhelp$Group.2,outhelp$Group.3),]
    out$Value <- outhelp$x
    metahelp <- "|Diurnal Cycle=N"
  } else {
    print("already daily data")
    out$Value <- x[,"Value"]
  }
  ## Write daily file
  #  out <- out[which(!is.na(out$Value)), ]
  #  if (nrow(out) > 0) {
  write_sef(out, outpath, meta["var"], meta["id"], meta["name"], 
            meta["lat"], meta["lon"], meta["alt"], meta["source"], meta["link"], 
            meta["units"], "mean", period="day", paste0(meta["meta"],metahelp), note=new_note, keep_na=TRUE)
  #  }
}
nc_close(ncp)

########################
## Precipitation
########################

seffiles <- list.files(sef)
seffiles_rr <- seffiles[grepl("_rr.tsv|_rr_",seffiles)]

for (stn in seffiles_rr) {
  print(stn)
  addvar <- gsub(".tsv","",gsub(".*_rr_|.*_rr.tsv","", stn))
  new_note <- paste0(addvar,ifelse(nchar(addvar)>0,"_",""),"daily")
  x <- read_sef(paste0(sef,stn), all=TRUE)
  ko <- grep("qc", x$Meta)
  if (length(ko) > 0) x <- x[-ko, ] # Remove flagged values
  ko <- grep("cumulated", x$Meta)
  if (length(ko) > 0) x <- x[-ko, ] # Remove cumulated values
  ko <- grep("7day", x$Period)
  if (length(ko) > 0) x <- x[-ko, ] # Remove weekly values
  xsum <- aggregate(x$Value, list(x$Year,x$Month,x$Day), sum)
  xsum <- xsum[order(xsum$Group.1, xsum$Group.2, xsum$Group.3), ]
  x$Time <- x$Hour + x$Minute/60
  xtime <- aggregate(x$Time, list(x$Year,x$Month,x$Day), max)
  meta <- read_meta(paste0(sef,stn))
  
  
  ## Write daily file
  out <- data.frame(Year=xsum$Group.1,
                    Month=xsum$Group.2,
                    Day=xsum$Group.3,
                    Hour=trunc(xtime$x),
                    Minute=round((xtime$x-trunc(xtime$x))*60, 0),
                    Value=round(xsum$x, 1))
  #  out <- out[which(!is.na(out$Value)), ]
  write_sef(out, outpath, meta["var"], meta["id"], meta["name"],
            meta["lat"], meta["lon"], meta["alt"], meta["source"], meta["link"],
            meta["units"], "sum", period="p1day", meta["meta"], note=new_note, keep_na=TRUE)
}


rm(list=ls())
