# 020 Preprocessing
# preprocessing of input data for analogue reconstructions
# 2022-07-29


### input needed:
# specifications from masterscript
# data frame named TOT with first col "date" (date format), second col "weather type"
# and station data with colnames paste("3-LETTER-ABBREVIATION_","ta","p","rr","rr0,sep="")

#############################################################################################################
#############################################################################################################
# 1. PRESELECTION (stations, variables)
# 2. TREND CORRECTION (temperature)
# 3. SEASONALITY CORRECTION (temperature)
# 4. CLIMATE OFFSET FOR REFERENCE STATIONS
# 5. SEASONALITY CORRECTION (temperature)
# 6. STANDARDISE DATA
# 7. CONVERT RR0 DATA TO LOGICAL IN CASE OF GOWER AS DISTANCE METRIC
# 8. CREATE DATA FRAMES FOR PREDICTOR AND PREDICTAND
#############################################################################################################
#############################################################################################################

### load data
(load("data_input/analogue_stationdata.RData"))
### read inventory
inventory <- readWorksheetFromFile("data_input/arm_stations_inventory_withnbcn.xlsx",
                                   sheet=1, colTypes="character",startRow=2)

#############################################################################################################
# 1. PRESELECTION (stations, variables, weather types)
#############################################################################################################
print("select stations and variables")

data <- TOT
selcols <- which(colnames(data) %in% scenarios[[specifications$scenario]])
cols <- c(grep("WT|date",colnames(data)),selcols)
data <- data[,cols]

#############################################################################################################
# 2. TREND CORRECTION (temperature)
#############################################################################################################
if ("ta"%in%specifications$variables) {
  if (specifications$trend!=F) {
    print("removing temperature trend")
    
    t_dates <- data$date
    t_data <- data[,grepl("_ta",colnames(data))]
    
    ## coordinates of stations
    ids <- colnames(t_data)
    coords_ta <- as.numeric(inventory[match(ids,inventory$ID),c("lat_hist")])
    
    ## load zonal trend
    (load("data_input/era5_zonmean_trend_1961-2020.RData"))
    
    ## select ref period
    refind <- t_dates >= specifications$startRef
    t_data <- t_data[refind,]
    
    ## here, take the closest latitude for every station + select zonal mean trend
    nearlats <- sapply(coords_ta, function(x){which.min(abs(x-tr_list$lat))})
    temp <- t_data - t(tr_list$trend[nearlats,])
    
    data[refind,grepl("_ta",colnames(data))] <- temp
    
    rm(temp,t_data,t_dates,coords_ta, ids, refind,tr_list,nearlats)
    TOT <- data
    save(TOT, file=paste0("data_input/analogue_stationdata_detrended_refperiod_",scen,".RData"))
  } else {
    print("save station data no detrending")
    TOT <- data
    save(TOT, file=paste0("data_input/analogue_stationdata_",scen,".RData"))
  }

#############################################################################################################
# 4. CLIMATE OFFSET FOR REFERENCE STATIONS
#############################################################################################################
  if(specifications$offset==T){
  
  print("substract offset")
  
  t_dates <- data$date
  t_data <- data[,grepl("_ta",colnames(data))]
  
  ### load zonal offset
  offset <- read.table("data_input/runclim_offset_year.txt", header=T)
  histind <- data$date < specifications$startRef
  t_data_hist <- t_data[histind,]
  offset_mat <- matrix(rep(offset[,2], ncol(t_data_hist)),ncol=ncol(t_data_hist), byrow = F)
  yind <- match(year(t_dates[histind]), offset[,1])
  temp_off <- t_data_hist +  offset_mat[yind,]
  
  data[histind,grepl("_ta",colnames(data))] <- temp_off
  
  rm(offset,offset_mat, yind, temp_off, histind, t_dates,t_data_hist)
  
}
  

#############################################################################################################
# 5. SEASONALITY CORRECTION (temperature)
#############################################################################################################
 if (specifications$seasonality==T) {
   
   print("filtering temperature seasonality")
   
   if(specifications$sep_ref_hist==T){
     
     ind <- list(ref = data$date >= specifications$startRef,
                 hist = data$date <= paste0(specifications$endHist,"-01-01"))
     
     for(per in 1:2){
       t_date <- data$date[ind[[per]]]
       t_data <- data[ind[[per]],grepl("ta",colnames(data))]
       # calculate seasonality-corrected temperature data
       deseas <- apply(t_data,2,rm_seas, dates=t_date)
       data[ind[[per]],colnames(data)%in%colnames(deseas)] <- deseas
       rm(deseas,t_data,t_date)
     }
     
   } else {
     t_date <- data$date
     t_data <- data[,grepl("ta",colnames(data))]
     
     # calculate seasonality-corrected temperature data
     deseas <- apply(t_data,2,rm_seas, dates=t_date)
     data[,colnames(data)%in%colnames(deseas)] <- deseas
     
     rm(deseas,t_data,t_date)
   }
   
   TOT <- data
 }
}


#############################################################################################################
# 6. STANDARDISE DATA
#############################################################################################################
if(specifications$standardize==T) {
  print("standardizing data")
  selcols <- gsub(".*_","",colnames(data)) %in% c("p","ta")
  
  if(specifications$sep_ref_hist==T){
    
    ind <- list(ref = data$date >= specifications$startRef,
                hist = data$date <= paste0(specifications$endHist,"-01-01"))
    
    for(per in 1:2){
      data_stand <- data[ind[[per]],selcols]
      data_stand <- data.frame(scale(data_stand))
      data[ind[[per]],colnames(data)%in%colnames(data_stand)]<-data_stand
    }
    rm(data_stand, per,ind)
    
  } else {
    data_stand <- data[selcols]
    data_stand <- data.frame(scale(data_stand))
    data[,colnames(data)%in%colnames(data_stand)]<-data_stand
    rm(data_stand)
  }
}


#############################################################################################################
# 7. CONVERT RR0 DATA TO LOGICAL IN CASE OF GOWER AS DISTANCE METRIC
#############################################################################################################

if(specifications$distance=="Gower"){
  if(any(grepl("rr0",colnames(data)))){
    
    dfrr0 <- as.data.frame(data[,grep("rr0",colnames(data))])
    aold <- ifelse(dfrr0[,1]==1, T,F)
      if(dim(dfrr0)[2]>1){
        for(i in 2:ncol(dfrr0)){
          a <- ifelse(dfrr0[,i]==1, T,F)
          aold <- cbind(aold,a)
        }
        rm(a)
      }
    data[,grep("rr0",colnames(data))] <- aold
    rm(aold,dfrr0)
  }
}

#############################################################################################################
# 8. CREATE DATA FRAMES FOR PREDICTOR AND PREDICTAND
#############################################################################################################

# ADD EQUAL PROBABILITIES TO WTS WHERE THEY ARE MISSING
naprob <- which(is.na(data$WT_probability1))[2:length(which(is.na(data$WT_probability1)))]
data[naprob,grepl("probability", colnames(data))] <- 1/7

# define period of predictor & predictand
predictordates <- seq(as.Date(specifications$startRef),as.Date(specifications$endRef),by="day")
predictor <- data[data$date%in%predictordates,!grepl("prob",colnames(data))]

# include the ref periods as well, if calonly is FALSE!
predictanddates <- switch(specifications$calonly+1,
                          unique(c(seq(as.Date(specifications$startCal[1]),as.Date(specifications$endCal[1]),by="day"), predictordates)),
                          c(seq(as.Date(specifications$startCal[1]),as.Date(specifications$endCal[1]),by="day")))
predictand <- data[data$date%in%predictanddates,]

print("preselection completed")

# remove variables
rm(predictordates,predictanddates,  selcols)


# END