### 053 EnKF
### Ensemble Kalman fitting (temperature)
### 2022-07-29

### input needed:
# analogue dates (with n best analogues); should be an ensemble of min 30 Analogues
# 3 temperature raster datasets (tabs, tabs_deseas, tabs_climatology)
# station measurements (TOT)
# station coordinates 
# monthly difference between station measurements and corresponding gridcell of TabsD
# alitutde and distance matrices calculated from previous scripts
# running offset, if offset set to TRUE
# observation error calculated from 051_obs_error.R, if not set to a constant

#############################################################################################################
# 1. READ DATA
#    1.1 READ ANALOGUE DATES
#    1.2 READ RASTER DATA
#    1.3 READ STATION DATA
# 2. KALMAN FITTING
#    2.1 DEFINE VALUES
#    2.2 CALCULATE DISTANCE MATRIX
#    2.3 Kalman Filter for each day in analogue dates
# 3. WRITE RESULTS TO FILES
#############################################################################################################
rm(list=ls())

require(raster)
require(lubridate)

### for testing purposes
# in_year = 2012
# scen = "scen3"
# date = "2022-09-06"
# startCal = "1961-01-01"
# endCal = "2020-12-31"
# L = 1000
# Z= 1000
# climoff = "T"
# dist = "Gower"

### the following lines take arguments (year) from console
ptm <- proc.time()

args <- commandArgs(TRUE)
if (length(args)>0) {
  in_year = as.numeric(args[1])
  scen = args[2]
  date = args[3]
  startCal = args[4]
  endCal = args[5]
  L = as.numeric(args[6])
  Z = as.numeric(args[7])
  climoff = args[8]
  dist = args[9]
  }


### functions
pad2 <- function(x, width=2){stringr::str_pad(x,width,pad="0")}
pdate <- function(dat){as.Date(paste0(dat[,1],"-",pad2(dat[,2]),"-",pad2(dat[,3])))}

print(in_year)
print(scen)
print(Z)
print(climoff)
detr <- T

### scenarios are saved in this list
source("C011_scenarios.R")

###################################################################################################
### KALMAN FITTING
###################################################################################################
# define setting for Kalman fitting
kalmanset <- list(
  analog_dates =        paste0("output/ARM_run_",date,"/",scen,"/",date,"_analog.dates_",startCal,"-",endCal,"_deseas",switch(detr+1,"_","_detrend_"),dist,scen,".txt"),
  station_data =        paste0("data_input/analogue_stationdata",switch(detr+1,"_","_detrended_refperiod_"),scen,".RData"),           # define station data (TOT)
  station_xy =          "data_input/arm_stations_inventory_withnbcn.xlsx", # define station coordinate file
  input_raster =        paste0("data_input/TabsD_ch01r.swiss.lv95",switch(detr+1,"_","_detrended_"),"19610101-20201231.nc"), 
  input_raster_deseas = paste0("data_input/TabsD_ch01r.swiss.lv95",switch(detr+1,"_","_detrended_"),"deseas_19610101-20201231.nc"), 
  climatology =         paste0("data_input/TabsD_ch01r.swiss.lv95",switch(detr+1,"_","_detrended_"),"clim_19610101-20201231.nc"),
  offset_file =         "data_input/runclim_offset_year.txt",
  offset = ifelse(climoff=="T",T,F),
  stationset = scen,
  localize = F,
  cutoff = L, # define horizontal cutoff distance L [km] (if localize = T)
  Z_cutoff = Z, # define vertical cutoff distance z [m]  (if localize = T)
  Rhist = paste0("data_input/obserror_hist_all_temp.tsv"), # define measurement error R [°K]
  Rhist2 = paste0("data_input/obserror_hist2_all_temp.tsv"), # define measurement error R [°K]
  Rref = paste0("data_input/obserror_ref_all_temp.tsv"), # define measurement error R [°K]
  add_stns = scenarios[[scen]][which(scenarios[[scen]] %in% c("MIL_ta","TOR_ta","HOH_ta","ROV_ta"))],
  wgs_coords = "data_input/TabsD_ch01r.swiss.lv95_202009280000_202009290000.nc"
)


if (!exists("in_year")) message("define year for which EnKF should be run")
stopifnot(exists("in_year"))

## create output folder for validation
rdir <- ifelse(grepl("obserror",kalmanset$Rref),"Robsest","R1")
loc <- ifelse(kalmanset$localize,paste0(kalmanset$cutoff,"km_",kalmanset$Z_cutoff,"m_",rdir,"_"),paste0(rdir,"_"))
subDir <- paste0("EnKF","_temp_",loc,Sys.Date())
dir.create(file.path(dirname(kalmanset$analog_dates),subDir), showWarnings = F)
outfolder <- paste(file.path(dirname(kalmanset$analog_dates),subDir),"/",sep="")
rm(subDir)


#############################################################################################################
### 1. READ DATA
#############################################################################################################

#----------------------------------------------------------------------------------------
### 1.1 READ ANALOGUE DATES
#----------------------------------------------------------------------------------------
### load analogue dates
analog.dates <- read.table(kalmanset$analog_dates,as.is=T)

#----------------------------------------------------------------------------------------
### 1.2 READ RASTER DATA
#----------------------------------------------------------------------------------------
### load input rasters
suppressWarnings(tabs_deseas <- brick(kalmanset$input_raster_deseas))
suppressWarnings(tabs_climatology <- brick(kalmanset$climatology))

if(in_year < 1961){
  offset <- read.table(kalmanset$offset_file, header = T)
}

suppressWarnings(CHlon <- brick(kalmanset$wgs_coords, varname="lon"))
suppressWarnings(CHlat <- brick(kalmanset$wgs_coords, varname="lat"))
climatology_vals <- rasterToPoints(tabs_climatology)

#----------------------------------------------------------------------------------------
### 1.3 READ STATION DATA
#----------------------------------------------------------------------------------------
if(scen=="scen5"){
  stationset <- c("GVE_ta","BAS_ta","GSB_ta","NEU_ta","BER_ta","CDF_ta","MER_ta","ENG_ta",
                  "LUZ_ta","SMA_ta","OTL_ta","LUG_ta","STG_ta","SAE_ta","RAG_ta","DAV_ta",
                  "SAM_ta","SIA_ta","SBE_ta")
} else{
  stationset <- scenarios[[kalmanset$stationset]]
}


### Read station temperature measurements
load(kalmanset$station_data) # this is called TOT
station_data <- data.frame("date"=as.Date(TOT$date),TOT[,stationset[grepl("_ta",stationset)]])
station_data_all <- station_data
stationset <- colnames(station_data[,-1])

#------------------------------------------------------------------------------------------------
## for console run
if (exists("in_year")) {
  station_data <- station_data[which(year(station_data$date)%in%in_year),]
  analog.dates <- analog.dates[which(substr(analog.dates$date,1,4)%in%in_year),]
}
#------------------------------------------------------------------------------------------------

### Read station coordinates
station_coords <- XLConnect::readWorksheetFromFile(kalmanset$station_xy,sheet=1, colTypes="character",startRow=2)[,c("ID","lon_ref","lat_ref","E_lv95_ref","N_lv95_ref")]
station_coords <- station_coords[match(stationset,station_coords$ID),]
station_coords$lon_ref <- as.numeric(station_coords$lon_ref)
station_coords$lat_ref <- as.numeric(station_coords$lat_ref)
station_coords$E_lv95_ref <- as.numeric(station_coords$E_lv95_ref)
station_coords$N_lv95_ref <- as.numeric(station_coords$N_lv95_ref)

### remove bias between temperature measurements and gridcells
### (mean bias calculated over the period 1961-2020 for each month)
diff <- read.table(paste0("data_input/diff_station_raster_all_temp.txt"), head=T)
id <- match(stationset,diff$ID)
diff <- diff[id,]
for(j in 2:ncol(station_data)){
  if(all(!is.na(diff[j-1,]))){
    for(i in c(1:12)){
      station_data[month(station_data$date)==i,j]<-station_data[month(station_data$date)==i,j]-diff[j-1,i+3]
    }
  }
}

### subset raster data
targetdays <- paste("X",gsub("-",".",analog.dates$date),sep="")
temp_kalman <- tabs_deseas[[1:nrow(analog.dates)]] ## just for defining the grid
names(temp_kalman) <- targetdays
temp_kalman_deseas <- temp_kalman

#############################################################################################################
### 2. KALMAN FITTING
#############################################################################################################

#----------------------------------------------------------------------------------------
### 2.1 DEFINE VALUES
#----------------------------------------------------------------------------------------
### define horizontal (L) and vertical (Z) localisation parameters (sigma in km (L) and m (Z))
if (kalmanset$localize == T)  {
  L = kalmanset$cutoff
  Z = kalmanset$Z_cutoff
}


### define measurement error
if(grepl("obserror",kalmanset$Rref)){
  suppressWarnings(Rref <- read.table(kalmanset$Rref, header=T))
  suppressWarnings(Rhist <- read.table(kalmanset$Rhist, header=T))
  suppressWarnings(Rhist2 <- read.table(kalmanset$Rhist2, header=T))
} else {
  Rref <- Rhist <- Rhist2 <- matrix(rep(kalmanset$Rref,length(station_coords$ID)), ncol=1)
  rownames(Rref) <- rownames(Rhist) <- rownames(Rhist2) <- station_coords$ID
}

#----------------------------------------------------------------------------------------
### 2.2 CALCULATE DISTANCE MATRIX
#----------------------------------------------------------------------------------------
if (kalmanset$localize ==T) {
  print("calculate distance matrix")
  ### read alt mat from code/data_input already based on scenario
  load(paste0("data_input/alt_mat_all_temp.RData"))
  ### read dist mat from code/data_input already based on scenario
  load(paste0("data_input/dist_mat_all_temp.RData"))
}

### get the lonlat values for each grid cell
values(CHlon[[1]])[is.na(values(tabs_climatology[[1]]))]<- NA
values(CHlat[[1]])[is.na(values(tabs_climatology[[1]]))]<- NA

lonvals <- rasterToPoints(CHlon)
latvals <- rasterToPoints(CHlat)
CHlonlat <- cbind(lonvals[,3],latvals[,3])


#----------------------------------------------------------------------------------------
### 2.3 Kalman Filter for each day in analogue dates
#----------------------------------------------------------------------------------------
#day <- 1

for (day in 1:nrow(analog.dates)){
  
  print(analog.dates$date[day])

  ### subset raster for analog days
  analogdays_date <- as.Date(as.numeric(analog.dates[day,-1]),origin = "1970-01-01")
  
  if(all(is.na(analogdays_date))){
    values(temp_kalman[[day]])[!is.na(values(temp_kalman[[day]]))]<- NA
    values(temp_kalman_deseas[[day]])[!is.na(values(temp_kalman_deseas[[day]]))]<- NA
  } else {
    
    analogdays <- paste("X",gsub("-",".",analogdays_date),sep="")
    
    temp_analogs_deseas <- subset(tabs_deseas,analogdays)
    
    ### define number of ensembles (evtl. change in kalman filter script)
    nens <- length(analogdays)
    doy <- paste("X2020.",gsub("-",".",substr(analog.dates$date[day],6,10)),sep="")
    
    ## change climatology depending on hist
    clim <- climatology_vals[,doy]
    
    if(in_year < 1961 & kalmanset$offset){
      print("remove offset")
      clim <- climatology_vals[,doy] - offset[offset$histyear == in_year,2]
    }
    
    ## create data matrix of analog days and add climatology
    rast_vals <- rasterToPoints(temp_analogs_deseas)
    rast_vals[,c(1:2)] <- CHlonlat
    rast_vals[,-c(1:2)] <- apply(rast_vals[,-c(1:2)],2,function(x) x+clim)
    
    ### subset station data and create observation vector
    station_data_sub <- station_data[station_data$date==analog.dates$date[day],]
    station_data_sub <- station_data_sub[,!is.na(station_data_sub)]
    
    ## order coordinates according to data
    stnnam <- colnames(station_data_sub[2:ncol(station_data_sub)])
    coords_sub <- station_coords[match(stnnam,station_coords$ID),] 
    obs <- station_data_sub[,2:ncol(station_data_sub)]
    
    ## add extra cells for outside station and estimate a model value
    if(length(kalmanset$add_stn)>0){
      nonch <- which(stnnam %in% kalmanset$add_stns)
      outside_vals <- matrix(NA, ncol=ncol(rast_vals), nrow=length(nonch))
      outside_vals[,1] <- coords_sub[nonch,2]
      outside_vals[,2] <- coords_sub[nonch,3]
      seldat <- match(analogdays_date,station_data_all$date)
      outside_vals[,3:ncol(outside_vals)] <- t(station_data_all[seldat,match(stnnam[nonch],colnames(station_data_all))]) ## same order as above
      colnames(outside_vals) <- c("x","y",as.character(analogdays_date ))
      rast_vals <- rbind(outside_vals,rast_vals)
    }
    
    ###### create H vector  ###### 
    stat <- cellFromXY(temp_analogs_deseas[[1]], cbind(unlist(coords_sub[,4]), unlist(coords_sub[,5])))
    cell.nrs <- which(!is.na(values(temp_analogs_deseas[[1]])))
    
    if(length(kalmanset$add_stn)>0){
      h.i <- match(stat,cell.nrs) + nrow(outside_vals) ## add the additional cells from outside_vals
      names(h.i) <- coords_sub$ID
      ## add positions from additional values
      h.i[stnnam[nonch]] <- match(stnnam[nonch], kalmanset$add_stns)
      obsind <- !is.na(h.i)
    } else {
      h.i <- match(stat,cell.nrs)  
      names(h.i) <- coords_sub$ID
      obsind <- !is.na(h.i)
    }
    
    h.i <-  h.i[obsind]
    H <- t(sapply(h.i, function(i) {out <- rep(0, nrow(rast_vals)) ; out[i] <- 1; out}))
    
    ### create observation covariance matrix
    R <- switch((in_year > 1863)+1,Rhist, switch((in_year >= 1961)+1,Rhist2, Rref))
    Rcal <- R[names(h.i),]
    
    ### define background mean and variance
    xb <- rast_vals[,-c(1:2)]
    xb_mean <- rowMeans(xb)
    xb_1 <- xb - xb_mean
    
    ### calculate covariance matrix
    print("calculate covariance matrix")
    
    PH <- matrix(ncol=length(h.i),nrow = nrow(rast_vals))
    if(kalmanset$localize==T) {
      for (j in seq_along(h.i)){
        mind <- which(names(h.i)[j] == rownames(dist_mat))
        exp_alt <- exp(-((abs(alt_mat[mind,])^2)/(2*Z^2)))
        exp_dist <- exp(-((abs(dist_mat[mind,])^2)/(2*L^2)))
         
        ###  corrtest
        # pcorr_mat <- sapply(1:nrow(xb_1), function(y){ as.numeric(cor.test(xb_1[h.i[j],],xb_1[y,])$estimate)})
        # pcorr_mat[pcorr_mat<0.2] <- 0
        # pcorr_mat[pcorr_mat>=0.2] <- 1
        # PH[,j] <- (xb_1 %*% t(xb_1[h.i[j],,drop=F]) / (nens - 1))*exp_alt*exp_dist*pcorr_mat
        # PHtest[,j] <- (xb_1 %*% t(xb_1[h.i[j],,drop=F]) / (nens - 1))*exp_alt*exp_dist
        
      }
    } else {
      for (j in seq_along(h.i)){
        PH[,j] <- (xb_1 %*% t(xb_1[h.i[j],,drop=F]) / (nens - 1))
      }
    }
    
    ### do Kalman fitting
    print("apply Kalman fitting")
    HPHR <- (H %*% PH + diag(Rcal))
    K <- PH %*% solve(HPHR)
    eHPHR <- eigen(HPHR)
    sqrtHPHR <- eHPHR$vectors %*% diag(sqrt(eHPHR$values)) %*% t(eHPHR$vectors)
    Ktilde <- PH %*% t(solve(sqrtHPHR)) %*% solve(sqrtHPHR + diag(sqrt(Rcal)))
    
    ## update for mean
    xa <- as.vector(xb_mean + K %*% (as.numeric(obs[obsind]) - (H %*% xb_mean)))
    
    ## update for anomaly
    xb_1_small <- xb_1[h.i,]
    a_small <- Ktilde %*% H[,h.i]     #[erstes Matrixprodukt]
    b <- a_small %*% xb_1_small       #[zweites Matrixprodukt]
    xa1 <- xb_1 - b
    analysis <- xa + xa1
    
    if(length(kalmanset$add_stn)>0){
      analysis <- analysis[(nrow(outside_vals)+1):nrow(analysis),] ## remove the added cells again
    }
    
    ### write data to raster files (temperature & deseas temperature)
    analysis_deseas <- apply(analysis,2,function(x) x-clim)
    values(temp_kalman[[day]])[!is.na(values(temp_kalman[[day]]))]<- as.numeric(analysis[,1])
    values(temp_kalman_deseas[[day]])[!is.na(values(temp_kalman_deseas[[day]]))]<- as.numeric(analysis_deseas[,1])
    
  }
  
}


#############################################################################################################
### 3. WRITE RESULTS TO FILES
#############################################################################################################
sdate <- as.character(analog.dates[1,1])
edate <- as.character(analog.dates[nrow(analog.dates),1])

if (kalmanset$localize==T) {
  props1 <- "_localized"
  props2 <- paste("_",kalmanset$cutoff,"km",sep="")
  props3 <- paste("_",kalmanset$Z_cutoff,"m",sep="")
} else {
  props1 <- c()
  props2 <- c()
  props3 <- c()
}

# set Z dimension to time
temp_kalman <- setZ(temp_kalman,seq(as.Date(sdate),as.Date(edate),by="day"),"time")
temp_kalman_deseas <- setZ(temp_kalman_deseas,seq(as.Date(sdate),as.Date(edate),by="day"),"time")

# write gridded files
print("write raster: temp_EnKF")
writeRaster(temp_kalman,file=paste(outfolder,Sys.Date(),"_temp_EnKF_",sdate,"-",edate,props1,props2,props3,".nc",sep=""),
            overwrite=T, varname="temp", varunit="degC",longname="daily mean temperature",zname="time",xname="E",yname="N")
print("write raster: temp_EnKF deseasonalised")
writeRaster(temp_kalman_deseas,file=paste(outfolder,Sys.Date(),"_temp_EnKF_deseas_",sdate,"-",edate,props1,props2,props3,".nc",sep=""),
            overwrite=T, varname="temp", varunit="degC",longname="daily mean temperature anomaly",zname="time",xname="E",yname="N")

readme<-data.frame("description"=c("kalman filter","period","analog dates","station data","station coordinates",
                                   "input raster: raw","input raster: deseas","input raster: offset","input raster: offset deseas",
                                   "input raster: clim","input raster: offset clim",
                                   "station set","localize","L [if localize ==T]","Z [if localize ==T]","R"),"parameters"=NA)

readme<-data.frame("description"=c("kalman filter","period","analog dates","station data","station coordinates",
                                   "input raster: raw","input raster: deseas","input raster: clim","offset:",
                                   "station set","localize","L [if localize ==T]","Z [if localize ==T]","R"),"parameters"=NA)

readme[1,2]<-as.character(as.Date(Sys.Date()))
readme[2,2]<-paste(sdate,edate,sep="-")
for(i in 3:nrow(readme)) {readme[i,2]<-as.character((paste(kalmanset[[i-2]], collapse = ", ")))}
readme[9,2]<-kalmanset$offset
readme[10,2]<-paste(stationset,collapse = ", ")
readme[11,2]<-kalmanset$localize
readme[12,2]<-kalmanset$cutoff
readme[13,2]<-kalmanset$Z_cutoff
readme[14,2]<-kalmanset$Rhist
write.table(readme,file=paste(outfolder,Sys.Date(),"_readme_EnKF_",sdate,"-",edate,props1,props2,props3,".txt",sep=""))

mtm <- proc.time()
print(paste("calculation time EnKF for ", in_year,":",(mtm[3]-ptm[3])%/%60,"minutes", round((mtm[3]-ptm[3])%%60),"seconds" ,sep=" "))
ptm <- mtm

# remove variables from environment
if(kalmanset$localize==T){rm(alt_mat,dist_mat,L,Z)}

# END

