#!/anaconda3/envs/my_env/bin/python

# -----------------------------------------------
# validation of temperature
# 29.04.2022
# -----------------------------------------------

# import packages 
import numpy as np
import pandas as pd
import xarray as xr
import dask
import bottleneck
from dask.distributed import Client
from datetime import datetime

import sys
import glob
import time
import warnings  # to omit warnings for spatial correlation (as division by zero (no rain days) is not possible)
import os

client = Client(n_workers=2, threads_per_worker=4, memory_limit='10GB')
#client

# ------------------------------------------------
# define functions
# ------------------------------------------------

rmse  = lambda x,y: np.sqrt(np.mean((x-y)**2, axis = -1))
mae   = lambda x,y: np.mean(np.abs(x-y), axis = -1)
bias  = lambda x,y: y.mean(axis = -1) - x.mean(axis = -1) 
msess = lambda x,y: 1 - np.sum((x-y)**2, axis = -1)/np.sum((0-x)**2, axis = -1)

def covariance_gufunc(x, y):
    return ((x - x.mean(axis = -1, keepdims = True))
            * (y - y.mean(axis = -1, keepdims = True))).mean(axis = -1)

def pearson_correlation_gufunc(x, y):
    return covariance_gufunc(x, y) / (x.std(axis=-1) * y.std(axis=-1))

## define wrapper functions for apply_ufunc
def pearson_correlation(x, y, dim):
    return xr.apply_ufunc(
        pearson_correlation_gufunc, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def bias_calculation(x, y, dim):
    return xr.apply_ufunc(
        bias, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def rmse_calculation(x, y, dim):
    return xr.apply_ufunc(
        rmse, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def msess_calculation(x, y, dim):
    return xr.apply_ufunc(
        msess, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def mae_calculation(x, y, dim):
    return xr.apply_ufunc(
        mae, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])
        
        
### the script is called from the console!
sys.argv = ["", "/scratch3/noemi/wear/swiss_arm/code/output/ARM_run_2022-09-06/scen3/EnKF_temp_Robsest_2022-09-06/"]

print("start")

# omit warnings (see above)
warnings.filterwarnings("ignore")

# get start time
stm = time.time()

# check if file is valid
if len(sys.argv) < 2:
    sys.exit("indicate input file for validation")

if not os.path.isdir(str(sys.argv[1])):
    sys.exit("file path invalid")

# create validation output folder
indir = os.path.dirname(str(sys.argv[1]))
exp = indir.split('/')
outdir = indir + "/validation_" + time.strftime("%Y-%m-%d") + "/"
print(outdir)
if not os.path.isdir(outdir):
    try:
        os.mkdir(outdir)
    except OSError:
        print("")
    else:
        print("")       
        
        
# --------------------------------------------------------------------------------------------------------------------
# READ DATA
# --------------------------------------------------------------------------------------------------------------------
print("read observation data")

obs_ifolder = "/scratch3/noemi/wear/swiss_arm/code/data_input/"
obs_ifile = "TabsD_ch01r.swiss.lv95_detrended_deseas_19610101-20201231.nc"
temp_obs = xr.open_dataset(obs_ifolder + obs_ifile, chunks = dict(E = 5))
temp_obs = temp_obs["TabsD"]  # temp_obs to dataarray (not dataset)
temp_obs

print("read reconstruction data")
rec_ifile = str(sys.argv[1])
allfiles = glob.glob(f'{rec_ifile}*deseas_1961-01-01-2020-12-31*.nc')
temp_rec = xr.open_mfdataset(paths = allfiles, engine = "netcdf4", chunks = dict(E = 5)).chunk({'time':-1})
temp_rec = temp_rec["temp"]
temp_rec = temp_rec.sortby(temp_rec.N)
timesel = temp_rec.time >=  pd.to_datetime("1961-01-01", format='%Y-%m-%d')
temp_rec = temp_rec[timesel,:,:]
temp_rec

# --------------------------------------------------------------------------------------------------------------------
# Start validation
# --------------------------------------------------------------------------------------------------------------------

seasons = {"MAM": np.arange(3,6),"JJA":np.arange(6,9),"SON":np.arange(9,12),"DJF": [12,1,2],"annual":np.arange(1,13)}
validation = []


for name,value in seasons.items():
    print(value)
    print(name)
    
    obs_seas = temp_obs.sel(time=np.isin(temp_obs.time.dt.month,value))#[:,100:110,100:110]
    rec_seas = temp_rec.sel(time=np.isin(temp_rec.time.dt.month,value))#[:,100:110,100:110]
    print("pcorr")
    pcorr_out = pearson_correlation(obs_seas, rec_seas, "time").compute()
    print("bias")
    bias_out = bias_calculation(obs_seas, rec_seas, "time").compute()
    print("rmse")
    rmse_out = rmse_calculation(obs_seas, rec_seas, "time").compute()
    print("mae")
    mae_out = mae_calculation(obs_seas, rec_seas, "time").compute()
    print("msess")
    msess_out = msess_calculation(obs_seas, rec_seas, "time").compute()
    
    metrics = xr.concat([pcorr_out, bias_out, rmse_out, mae_out, msess_out], dim ="metrics", fill_value = np.nan)
    newcoords = ["corr_" + name, "bias_"  + name, "rmse_"  + name, "mae_"  + name, "msess_"  + name]    
    metrics = metrics.assign_coords(metrics=newcoords)
    metrics.E.attrs['axis'] = "X"
    metrics.E.attrs['units'] = "degrees_east"
    metrics.N.attrs['axis'] = "Y"
    metrics.N.attrs['units'] = "degrees_north"
    
    validation.append(metrics)
    
# --------------------------------------------------------------------------------------------------------------------
# Write file
# --------------------------------------------------------------------------------------------------------------------
    
print("write netcdf")
ds = xr.concat([validation[0],validation[1],validation[2],validation[3],validation[4]], dim="metrics")
ds = ds.rename("vals")
ds.to_netcdf(outdir + time.strftime("%Y-%m-%d") + '_validation_maps_temp.nc')

print("end")    
    
    

